/*
 * Copyright (C) 2002-2012 OpenClovis Solutions Inc.  All Rights Reserved.
 *
 * This file is available  under  a  commercial  license  from  the
 * copyright  holder or the GNU General Public License Version 2.0.
 * 
 * The source code for  this program is not published  or otherwise 
 * divested of  its trade secrets, irrespective  of  what  has been 
 * deposited with the U.S. Copyright office.
 * 
 * This program is distributed in the  hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied  warranty  of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 * 
 * For more  information, see  the file  COPYING provided with this
 * material.
 */
/*******************************************************************************
 * ModuleName  : cm
 * File        : clCmDefs.h
 *******************************************************************************/

/*******************************************************************************
 * Description :
 * This private header contains the enumerations and definitions internal 
 * to the chassis manager module                                           
 **************************************************************************/


#ifndef _CM_DEFS_H 
#define _CM_DEFS_H 

#ifdef __cplusplus 
extern "C" { 
#endif 

#include <stdio.h>

#include <SaHpi.h>    
//#include <oh_utils.h>

#include <clCommon.h>
#include <clEoApi.h>
#include <clOsalApi.h>
#include <clChassisMgrApi.h>
#include <clEventApi.h>
#include <clCpmApi.h>
#include <ipi/clCmIpi.h>

#include <clVersionApi.h>

/**
 * Return code generator macro for CM related errors
 */
#define CL_CM_ERROR(error)              (CL_RC(CL_CID_CM, error))


#define CL_CM_INVALID_PHYSICAL_SLOT 0xff 

/* Default values for auto-extract and aut-insert timeouts */
#define CL_CM_MIN_AUTOINSERT_TIMEOUT   (5*1000000000LL) /* 5 seconds */
#define CL_CM_MIN_AUTOEXTRACT_TIMEOUT  (0*1000000000LL) /* 0 seconds */

/* Name of the component registered with Debug Server */
#define CLOVIS_CHASSIS_MANAGER  "ClChassisManager"
#define INVALID_ALARM_ID 0x0 
#define CL_CHASSIS_MANAGER_RMDCALL_TIMEOUT CL_RMD_DEFAULT_TIMEOUT    
#define CL_CHASSIS_MANAGER_RMDCALL_RETRIES CL_RMD_DEFAULT_RETRIES    


/* EO Client side function enumerations */ 
#define  CL_CHASSIS_MANAGER_FRU_STATE_GET CL_EO_GET_FULL_FN_NUM(CL_EO_NATIVE_COMPONENT_TABLE_ID, 0x01)
#define  CL_CHASSIS_MANAGER_OPERATION_REQUEST CL_EO_GET_FULL_FN_NUM(CL_EO_NATIVE_COMPONENT_TABLE_ID, 0x02)
#define  CL_CHASSIS_MANAGER_CPM_RESPONSE_HANDLE CL_EO_GET_FULL_FN_NUM(CL_EO_NATIVE_COMPONENT_TABLE_ID, 0x03)
#define  CL_CHASSIS_MANAGER_THRESHOLD_STATE_GET CL_EO_GET_FULL_FN_NUM(CL_EO_NATIVE_COMPONENT_TABLE_ID, 0x04)

/* Configurable option for MULTIPLE CHASSIS */

extern ClUint16T clCmMaxChassisBits;
extern ClUint16T clCmMaxPhysicalBits;
extern ClUint16T clCmMaxAMCBits;

extern ClUint16T  clCmMaxChassis;
extern ClUint16T  clCmMaxPhysicalSlots;
extern ClUint16T  clCmMaxAMCSlots;

#define  CL_CM_MAX_NODE_BITS    (12)   /* With tipc we are supoorting 4096 address */
#define  CHASSIS_FACTOR (clCmMaxChassisBits?(1 << (CL_CM_MAX_NODE_BITS - clCmMaxChassisBits)):0)
#define  CHASSIS_ID_MASK (clCmMaxChassisBits?((1 << (clCmMaxChassisBits))-1):0)
#define  PHYSICAL_SLOT_FACTOR  (1 << ( CL_CM_MAX_NODE_BITS - clCmMaxChassisBits - clCmMaxPhysicalBits))
#define  AMC_FACTOR       (clCmMaxAMCBits?1:0)
#define  PHYSICAL_SLOT_MASk  ((1<<clCmMaxPhysicalBits)-1)

/* Event Buffer Definitions */
typedef struct clCmEventBuffer clCmEventBufferT;

typedef struct ClCmEventBufferDomain
{
    SaHpiDomainIdT    domainId;
    clCmEventBufferT  *clCmEventQueueHead;
    clCmEventBufferT  *clCmEventQueueTail;
    ClUint32T         clCmEventQueueNodeCount;
    ClOsalMutexT      eventQueueMutex;
    ClUint32T         peerHpiLastCount;
    ClListHeadT       next;
}ClCmEventBufferDomainT;

struct clCmEventBuffer
{
    SaHpiDomainIdT  domainId;
    SaHpiSessionIdT sessionId;
    SaHpiEventT event;
    SaHpiRptEntryT rptentry;
    SaHpiRdrT rdr;
    ClRcT pluginOps;
    ClCmEventBufferDomainT **domainQueueInfo;
    clCmEventBufferT *next;
};

typedef struct
{
   SaHpiSessionIdT  sessionId;
   SaHpiDomainIdT   domainId;
   ClOsalTaskIdT    eventListenerThread;
   ClBoolT          stopEventProcessing;
   ClUint8T         discoveryInProgress;
   ClListHeadT      list;
}ClCmDomainListT;

/* Chassis manager operation request structures */
typedef struct
ClCmOpReq
{
    ClVersionT      version;
    ClUint32T       chassisId;
    ClUint32T       physicalSlot;
    ClCmFruOperationT   request;

}ClCmOpReqT;    

typedef struct ClCmFruAddress
{
    ClVersionT      version;
    ClUint32T       chassisId;
    ClUint32T       physicalSlot;
}ClCmFruAddressT;    


/* Response structure from the CPMG */
typedef struct cpmResponse
{

    ClUint32T  seqNum;/* cookie to identify for which resource the query was made */
    ClUint8T  goAhead;/* Boolean response from CPMG whether to goahead with extraction or
                         insertion of the Card */
}cpmResponseT;

typedef struct ClCmThresholdStateReq
{
    ClVersionT version;
    ClUint32T slot;
    SaHpiSensorNumT sensorNum;
}ClCmThresholdStateReqT;

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* Event interface Information                                               */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
typedef struct ClCmEvtIface
{

    ClUint8T              eventIfaceInitDone;    /* CL_TRUE or CL_FALSE */ 
    ClEventHandleT        EvtHandle;        /* Handle to the Event */
    ClEventChannelHandleT EventChannelHandle;/* Handle to the CM Event Channel
                                                */
    ClEventInitHandleT    EventLibHandle;    /*Handle to the Event Library */

}ClCmEvtIfaceInfoT;

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* Platform related information                                           */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
#define CL_CM_MAX_HPI_RESOURCES_PER_DOMAIN 64 
typedef struct ClCmPlatformInfo 
{

    ClUint32T no_of_sensors;
    ClUint32T no_of_inventories;
    ClUint32T no_of_frus;
    SaHpiDomainIdT  defaultDomainId;
    SaHpiSessionIdT hpiSessionId; /* session id of the first session */
    ClUint8T        stopEventProcessing; /* switch for turning HPI Event
                                            processing CLTRUE or CL_FALSE */
    ClUint8T         discoveryInProgress; /*CL_TRUE if discovery in progress */
    SaHpiResourceIdT fruResourceIds[CL_CM_MAX_HPI_RESOURCES_PER_DOMAIN];
}ClCmPlatformInfoT;

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/* Chassis manager Context Information                                         */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/

typedef struct ClCmContext
{
    ClEoExecutionObjT* executionObjHandle; /* Handle to the ChassisManager 
                                              execution Object */
    ClCpmHandleT       cmCompHandle; /* Handle to the Chassis manager component */
    ClUint8T            cmInitDone;  /* CL_TRUE or CL_FALSE */     
    ClCmPlatformInfoT  platformInfo; /* platform related information */
    ClCmEvtIfaceInfoT  eventInfo;    /* event interface information  */
    ClOsalTaskIdT   eventListenerThread ;
}ClCmContextT;    


ClRcT VDECL(clCmServerApiFruStateGet)( ClEoDataT ,
                                       ClBufferHandleT,
                                       ClBufferHandleT);

ClRcT VDECL(clCmServerOperationRequest)( ClEoDataT ,
                                         ClBufferHandleT, 
                                         ClBufferHandleT);

ClRcT VDECL(clCmServerCpmResponseHandle)(ClEoDataT,
                                         ClBufferHandleT,
                                         ClBufferHandleT);

ClRcT VDECL(clCmServerApiThresholdStateGet)(ClEoDataT,
                                            ClBufferHandleT,
                                            ClBufferHandleT);

/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
/*    information of waiting FRUs for CPMG's response                       */
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/


/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/




ClRcT cmDebugRegister( ClEoExecutionObjT *peoObj );

ClRcT cmDebugDeregister( ClEoExecutionObjT *peoObj );
/*------------------- CM HPI Interface funcs -----------------*/

extern ClRcT clCmHpiInterfaceInitialize(void );

extern ClRcT clCmHpiInterfaceFinalize(void );



ClRcT clCmFullHotSwapHandler(SaHpiSessionIdT sessionid,
                             SaHpiEventT * pEvent,
                             SaHpiRptEntryT  * pRptEntry);

ClRcT clCmSimpleHotSwapHandler ( SaHpiSessionIdT sessionid , 
                                 SaHpiEventT * pEvent ,
                                 SaHpiRptEntryT  * pRptEntry);

ClRcT clCmHotSwapCpmUpdate( SaHpiRptEntryT * pRptEntry,
                            ClCmCpmMsgTypeT cm2CpmMsgType);
                            
void clCmHPICpmResponseHandle(ClCpmCmMsgT * pCpm2CmMsg);



ClRcT WalkThroughDomain  ( SaHpiDrtEntryT   *);

ClRcT WalkThroughResource (SaHpiRptEntryT* ,
                           SaHpiSessionIdT );



/************Events Published by CM *************/
typedef enum 
{
    CM_FRU_STATE_TRANSITION_EVENT =0,
    CM_ALARM_EVENT 
}CmEventIdT;



extern ClRcT clCmPublishEvent(CmEventIdT cmEventId ,ClUint32T payLoadLen,void * pEventData );
extern ClRcT clCmHpiPublishEvent(CmEventIdT cmEventId ,ClUint32T payLoadLen,void * pEventData );


/*-------------------------Utility Stuff ------------------------*/


extern ClUint32T clCmPhysicalSlotFromEntityPath( SaHpiEntityPathT * pEpath );

extern void cmPrint( char * , ... );


extern ClRcT clCmEventInterfaceInitialize(void );

extern ClRcT clCmEventInterfaceFinalize( void );


extern ClRcT clCmHandleFruHotSwapEvent( SaHpiSessionIdT sessionId, SaHpiHsStateT prevState , 
                                        SaHpiHsStateT presState , 
                                        SaHpiRptEntryT * pRptEntry );

extern ClRcT clCmSimpleHotSwapHandler ( SaHpiSessionIdT sessionid , 
                                 SaHpiEventT * pEvent ,
                                 SaHpiRptEntryT  * pRptEntry);

extern SaHpiResourceIdT clCmResIdFromEpath(SaHpiSessionIdT sessionId,SaHpiEntityPathT * pEpath);

extern  ClRcT clCmGetResourceIdFromSlotId(ClUint32T chassisId, ClUint32T physicalSlot, SaHpiResourceIdT *pResourceId, SaHpiSessionIdT *sessionId );


/*------------------------- Version Stuff ------------------------*/

/* CM Fru State operations structure with version field */ 
extern ClVersionDatabaseT gCmClientToServerVersionDb;

#define clCmClientToServerVersionValidate(version) \
		do\
		{ \
			if(clVersionVerify(&gCmClientToServerVersionDb, (&version) ) != CL_OK) \
			{ \
			    clLog(CL_LOG_SEV_ERROR, CL_LOG_AREA_UNSPECIFIED, CL_LOG_CONTEXT_UNSPECIFIED,"Failed to validate the client version ."); \
			    return CL_ERR_VERSION_MISMATCH;  \
			} \
		}while(0) 


#define CL_CM_VERSION_SET(version)     \
		do { 							\
			version.releaseCode = (ClUint8T)'B' ; \
			version.majorVersion = 0x1; \
			version.minorVersion = 0x1; \
		}while(0)

/************************Debugging Macros ************************/
#ifdef CL_CM_CLIENT_DEBUG     
    #define CL_CM_CLIENT_DEBUG_PRINT(Z) \
    do{\
        printf("\n File=%s,Function=%s,Line=%d ",__FILE__,__FUNCTION__,__LINE__);\
        printf Z;\
        printf("\n");\
      }while(0)
#else
    #define CL_CM_CLIENT_DEBUG_PRINT(Z) 
#endif
    
#if 0
#ifdef CL_CM_SERVER_DEBUG     
    #define CL_CM_SERVER_DEBUG_PRINT(Z) \
    do{\
        printf("\n File=%s,Function=%s,Line=%d ",__FILE__,__FUNCTION__,__LINE__);\
        printf Z;\
        printf("\n");\
      }while(0)
#else
    #define CL_CM_SERVER_DEBUG_PRINT(Z) 
#endif
   
#ifdef CL_CM_HS_DEBUG     
    #define CL_CM_HS_DEBUG_PRINT(Z) \
    do{\
        printf("\n File=%s,Function=%s,Line=%d ",__FILE__,__FUNCTION__,__LINE__);\
        printf Z;\
        printf("\n");\
      }while(0)
#else
    #define CL_CM_HS_DEBUG_PRINT(Z) 
#endif
#endif

typedef struct
{
   SaHpiDomainIdT   domainId;
   ClInt8T          hpiConnectionStatus;
   ClUint32T        hpiConnectionCount;
}ClCmDomainHpiStatusT;

#define CL_MAX_DOMAINS                10
#define CL_CM_FALSE                   0X00
#define CL_CM_TRUE                    0x01
#define CL_CM_TRY_AGAIN               0x02

#define CL_CM_MAX_EVENT_QUEUE_SIZE    5000

#define CL_CM_HPI_CONNECTION_UP                      1
#define CL_CM_HPI_CONNECTION_DOWN                    0
#define CL_CM_HPI_CONNECTION_POLL_TIME               1 //sec

#define CL_CM_HPI_CONNECTION_VAL        (ClUint32T) 0xFFFFFFFF

#ifdef __cplusplus 
}
#endif 
#endif  /* end #ifndef _CM_DEFS_H */ 
