/*
 * Copyright (C) 2002-2009 by OpenClovis Inc. All  Rights Reserved.
 * 
 * The source code for  this program is not published  or otherwise 
 * divested of  its trade secrets, irrespective  of  what  has been 
 * deposited with the U.S. Copyright office.
 * 
 * This program is  free software; you can redistribute it and / or
 * modify  it under  the  terms  of  the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * 
 * This program is distributed in the  hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied  warranty  of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 * 
 * You  should  have  received  a  copy of  the  GNU General Public
 * License along  with  this program. If  not,  write  to  the 
 * Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*
 * Build: 4.2.0
 */
/*******************************************************************************
 * ModuleName  : cm                                                            
 * File        : clCmDbgCli.c
 *******************************************************************************/

/*******************************************************************************
 * Description :                                                                
 * This file contains the implementation of Chassis manager debug cli     
 * commands.                                                                  
 *******************************************************************************/



#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include <clCommon.h>
#include <clCommonErrors.h>
#include <clCmDefs.h>
#include <clDebugApi.h>
#include <clCmHpiDebug.h>

#include "clCmUtils.h"

#define AREA_CLI "CLI"
extern ClListHeadT gClDomainList;
extern ClOsalMutexT gClDomainListLock;

static ClRcT fakeResourceHotswap ( int argc , char **argv , char **retStr );
static ClRcT cmCliHotSwapEvent(int argc, char **argv, char **retStr);
static ClRcT showDomainsInfo(int argc, char **argv, char **retStr);

extern void cmCliPrint( char **retstr, char *format , ... );
extern ClCmContextT gClCmChassisMgrContext;
extern ClCmDomainListT* clCmDomainEntryGet( SaHpiDomainIdT domainId);
extern ClRcT clCmDomainIdFind( SaHpiDomainIdT domainId);
ClHandleT  gCmDebugReg  = CL_HANDLE_INVALID_VALUE;

static ClDebugFuncEntryT cmDbgCliCallbacks [] = 
{
    { (ClDebugCallbackT)fakeResourceHotswap ,
        "hotswap",
        "Used to simulate the extraction and insertion of the card by s/w"
    },

    {
        (ClDebugCallbackT) cmCliHotSwapEvent,
        "cmHotSwapEvent",
        "Simulate various hot swap events"
    },
    
    {
       (ClDebugCallbackT) showDomainsInfo,
        "domainsInfo",
        " Display domain related Information"
    }
};



/* Register the chassis manager with the Debug server */ 
ClRcT cmDebugRegister( ClEoExecutionObjT *peoObj )
{
  
    ClRcT  rc = CL_OK;

    rc = clDebugPromptSet("cm");
    if( CL_OK != rc )
    {
	return rc;
    }
    return clDebugRegister( cmDbgCliCallbacks,
                            (ClUint32T)(sizeof(cmDbgCliCallbacks)/
                            sizeof(cmDbgCliCallbacks[0])), 
                            &gCmDebugReg);
}


/* Deregister the Chassis manager with the Debug server */
ClRcT cmDebugDeregister( ClEoExecutionObjT *peoObj )
{
    return clDebugDeregister( gCmDebugReg ); 
}

ClDebugModEntryT clModTab[] =
{
	{ "cm", "cm", cmDbgCliCallbacks, "CM Commands"},
	{ "", "", 0, ""}
};
/*
   Debug cli commands to simulate the fake hotswap of the resource , should be
   only used for testing purpose only
*/
static ClRcT fakeResourceHotswap ( int argc , char **argv , char **retStr )
{

    ClRcT rc = CL_OK;
    SaHpiDomainIdT domainId;  
    SaHpiResourceIdT resourceId = 0x0;
    SaHpiHsActionT action = 0x0;
    ClCmDomainListT *domainListEntry = NULL;
    extern char *oh_lookup_error(SaErrorT );
    SaErrorT error = SA_OK;

    /* Allocate maximum possible */ 
    *retStr = clHeapAllocate(DISPLAY_BUFFER_LEN);
    if( *retStr == NULL )
    {
        clLog(CL_LOG_SEV_ERROR, AREA_CLI, CL_LOG_CONTEXT_UNSPECIFIED,
            "Could not allocate memory for CLI request");
        return CL_ERR_NO_MEMORY;
    }

    if( (argc !=3 ) && (argc != 4) )
    {
        cmCliPrint( retStr , "\nUsage: setstate resId  op=<ins or ext> \n"
                             "resId [DEC]- resourceid of the resource  "
                             "op [STRING]- operation can be ins or ext"
                             " DomainId - optional ");
        return rc;
    }
    
    resourceId = (SaHpiResourceIdT)atoi(argv[1]);

    if( strcmp(argv[2] , "ins" )== 0)
    {
        action = SAHPI_HS_ACTION_INSERTION;
    }
    else if (strcmp (argv[2] ,"ext")==0 )
    {
        action = SAHPI_HS_ACTION_EXTRACTION;
    }
    else
    {
        cmCliPrint( retStr, "\n Invalid state given ");
        return rc;
    }
    if( argc == 4)
    {
        domainId = (SaHpiDomainIdT)atoi(argv[3]);
        rc =  clCmDomainIdFind(domainId);
        if( rc != CL_OK)
        {
             cmCliPrint( retStr, "\n Invalid DomainID given ");
             return rc;
        }
    }
    else
    {
        domainId = gClCmChassisMgrContext.platformInfo.defaultDomainId;
    }
     
    domainListEntry = clCmDomainEntryGet(domainId);
    if( domainListEntry == NULL)
    {
         cmCliPrint( retStr, "\n Invalid DomainID given ");
         return rc;
    }
    
    if((error = saHpiHotSwapActionRequest(domainListEntry->sessionId, resourceId, action ))!= SA_OK )
    {
        cmCliPrint(retStr, "\nsaHpiHotSwapAction request failed :%s", oh_lookup_error(error));
        return rc;
    }
    clLog(CL_LOG_SEV_DEBUG, AREA_CLI,  CL_LOG_CONTEXT_UNSPECIFIED,
            "HS action requested on resource id [0x%x]: %d (%s)",
            resourceId, action, argv[2]);
    
    return rc;
}

static ClRcT cmCliHotSwapEvent(int argc, char **argv, char **retStr)
{
    ClRcT rc = CL_OK;
    
    /* Allocate maximum possible */ 
    *retStr = clHeapAllocate(DISPLAY_BUFFER_LEN);
    if( *retStr == NULL )
    {
        clLog(CL_LOG_SEV_ERROR, AREA_CLI, CL_LOG_CONTEXT_UNSPECIFIED,
            "Could not allocate memory for CLI request");
        return CL_ERR_NO_MEMORY;
    }

    if (argc != 3)
    {
        cmCliPrint(retStr,
                   "Usage : cmHotSwapEvent <node-address> <event-type-string>\n"
                   "Event type string should be one of the following :\n"
                   "1. surprise-extraction.\n"
                   "2. request-extraction.\n"
                   "3. request-insertion.\n"
                   "4. node-error-report.\n"
                   "5. node-error-clear.\n");
        return CL_ERR_INVALID_PARAMETER;
    }
    else
    {
        ClCmCpmMsgT cmCpmMsg;

        memset(&cmCpmMsg, 0, sizeof(ClCmCpmMsgT));

        cmCpmMsg.physicalSlot = atoi(argv[1]);

        if (!strcasecmp(argv[2], "surprise-extraction"))
            cmCpmMsg.cmCpmMsgType = CL_CM_BLADE_SURPRISE_EXTRACTION;
        else if (!strcasecmp(argv[2], "request-extraction"))
            cmCpmMsg.cmCpmMsgType = CL_CM_BLADE_REQ_EXTRACTION;
        else if (!strcasecmp(argv[2], "request-insertion"))
            cmCpmMsg.cmCpmMsgType = CL_CM_BLADE_REQ_INSERTION;
        else if (!strcasecmp(argv[2], "node-error-report"))
            cmCpmMsg.cmCpmMsgType = CL_CM_BLADE_NODE_ERROR_REPORT;
        else if (!strcasecmp(argv[2], "node-error-clear"))
            cmCpmMsg.cmCpmMsgType = CL_CM_BLADE_NODE_ERROR_CLEAR;

        rc = clCpmHotSwapEventHandle(&cmCpmMsg);
        if (CL_OK != rc)
        {
            cmCliPrint(retStr,
                       "CM hot swap event reporting failed, error [%#x]",
                       rc);
            return rc;
        }
    }
    return CL_OK;
}

char s[CL_MAX_NAME_LENGTH];

char * displayTime(SaHpiTimeT time)
{
   struct tm t1;
   time_t t;
   strcpy(s,"");
   if( time > SAHPI_TIME_MAX_RELATIVE)
   {
      t = (time_t)(time / 1000000000);
      localtime_r(&t,&t1);
      strftime(s, CL_MAX_NAME_LENGTH, "%m-%d-%Y %H-%M-%S",&t1);
   }
   else if(time == SAHPI_TIME_UNSPECIFIED)
   {
       strcpy(s, "SAHPI_TIME_UNSPECIFIED");
   }
   else if (time > SAHPI_TIME_UNSPECIFIED)
   {
       strcpy(s, "Invalid time ");
   }
   else
   {
      t = (time_t)(time / 1000000000);
      localtime_r(&t,&t1);
      strftime(s, SAHPI_MAX_TEXT_BUFFER_LENGTH, "%m-%d-%Y %H-%M-%S",&t1);
   }
   return s;
}
static ClRcT printDomainInfo(SaHpiSessionIdT sessionId,ClBufferHandleT msgHandle,ClBoolT verbose)
{
    SaErrorT ret   = SA_OK;
    SaHpiDomainInfoT domainInfo;
    ret = _saHpiDomainInfoGet(sessionId,&domainInfo);
    if( ret != SA_OK)
    {
         clDebugPrint(msgHandle, "saHpiDomainInfoGet failed, error [%#x]",ret);
         return CL_CM_ERROR(CL_ERR_CM_HPI_ERROR);
    }
    clDebugPrint(msgHandle,"  %-50s: %-8d \n", "Domain Id", domainInfo.DomainId);
    clDebugPrint(msgHandle,"  %-50s: %-8s \n", "Domain Name", domainInfo.DomainTag.Data);
    clDebugPrint(msgHandle,"  %-50s: 0x%-8x \n", "Domain Capability", domainInfo.DomainCapabilities);
    clDebugPrint(msgHandle,"  %-50s: %-8u \n", "Is Peer", domainInfo.IsPeer);
    if(verbose)
    {
        clDebugPrint(msgHandle,"  %-50s: %-8u \n", "Domain Refernce Table update count", domainInfo.DrtUpdateCount);
    	clDebugPrint(msgHandle,"  %-50s: %-8s \n", "DRT Timestamp", displayTime(domainInfo.DrtUpdateTimestamp));
    	clDebugPrint(msgHandle,"  %-50s: %-8u \n", "Resource Presence Table update count", domainInfo.RptUpdateCount);
    	clDebugPrint(msgHandle,"  %-50s: %-8s \n", "RPT Timestamp", displayTime(domainInfo.RptUpdateTimestamp));
    	clDebugPrint(msgHandle,"  %-50s: %-8u \n", "Domain Alarm Table update count", domainInfo.DatUpdateCount);
    	clDebugPrint(msgHandle,"  %-50s: %-8s \n", "DAT Timestamp", displayTime(domainInfo.DatUpdateTimestamp));
    	clDebugPrint(msgHandle,"       %-45s: %-8u \n", "ActiveAlarms", domainInfo.ActiveAlarms);
    	clDebugPrint(msgHandle,"       %-45s: %-8u \n", "CriticalAlarms", domainInfo.CriticalAlarms);
    	clDebugPrint(msgHandle,"       %-45s: %-8u \n", "MajorAlarms", domainInfo.MajorAlarms);
    	clDebugPrint(msgHandle,"       %-45s: %-8u \n", "MinorAlarms", domainInfo.MinorAlarms);
    }
    SaHpiEntryIdT    current, next;
    SaHpiDrtEntryT   drt_entry;
    current = SAHPI_FIRST_ENTRY;
    while( current != SAHPI_LAST_ENTRY )
    {
        if((ret = _saHpiDrtEntryGet(sessionId, current,  &next,  &drt_entry ))!= SA_OK )
        {
            if(ret == SA_ERR_HPI_NOT_PRESENT && current == SAHPI_FIRST_ENTRY)
            { 
                clDebugPrint(msgHandle,"  DRT is Empty\n");
                break;
            }
            return CL_CM_ERROR(CL_ERR_CM_HPI_ERROR);  
            
        }
        clDebugPrint(msgHandle,"  %-30s: %-8u \n", "Domain Entry in DRT is DomainId", drt_entry.DomainId);
        clDebugPrint(msgHandle,"  %-30s: %-8u \n", "IsPeer ", drt_entry.IsPeer);
        current = next;
    }
    return CL_OK;
}
   
     
static ClRcT showDomainsInfo(int argc, char **argv, char **retStr)
{

    ClRcT rc = CL_OK;
    ClListHeadT *iter = NULL;
    ClBufferHandleT msgHandle = 0;
    ClBoolT verbose;
    
    if( (rc = clDebugPrintInitialize(&msgHandle)) != CL_OK) 
        return rc;

    if( ( argc !=1 && argc !=2) ||  !retStr )
    {
        clDebugPrint(msgHandle,"Usage showDomainsInfo [-v] \n");
        rc = CL_RMD_RC(CL_ERR_INVALID_PARAMETER);
        goto out;
    }
     
    *retStr = NULL;

    if(argc == 2 && ( strcmp (argv[1],"-v") == 0))
    {
         verbose = CL_TRUE;
    }
    else
    {
         verbose = CL_FALSE;
    }        
    
    clOsalMutexLock(&gClDomainListLock);
    CL_LIST_FOR_EACH(iter, &gClDomainList)
    {
        ClCmDomainListT *entry = CL_LIST_ENTRY(iter,ClCmDomainListT,list);
        rc = printDomainInfo(entry->sessionId, msgHandle,verbose);
        if(rc != CL_OK)
        {
            clDebugPrint(msgHandle, "printDomainInfo returned with [%#x]\n",rc);
            goto out_free;
        }
        clDebugPrint(msgHandle,"\n"); 
    }
         
out_free:
    clOsalMutexUnlock(&gClDomainListLock);
out:
    clDebugPrintFinalize(&msgHandle, retStr); 
    return rc;     

}
