/*
 * Copyright (C) 2002-2009 by OpenClovis Inc. All  Rights Reserved.
 * 
 * The source code for  this program is not published  or otherwise 
 * divested of  its trade secrets, irrespective  of  what  has been 
 * deposited with the U.S. Copyright office.
 * 
 * This program is  free software; you can redistribute it and / or
 * modify  it under  the  terms  of  the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 * 
 * This program is distributed in the  hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied  warranty  of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 * 
 * You  should  have  received  a  copy of  the  GNU General Public
 * License along  with  this program. If  not,  write  to  the 
 * Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*
 * Build: 4.2.0
 */
/*******************************************************************************
 * ModuleName  : cm
 * File        : clCmEventInterface.c
 *******************************************************************************/

/*******************************************************************************
 * Description :
 * This file contains the interface implementation of the chassis manager 
 * with ASP.
 ****************************************************************************/

#include <SaHpi.h>
#include <clChassisMgrApi.h>
#include <clCommon.h>
#include <clDebugApi.h>
#include <clEventApi.h>
#include <clCmDefs.h>
#include <string.h>
#include <unistd.h>
#include <saEvt.h>
#include <clHpiEvtInterface.h>

#define AREA_EVT "EVT" /* log area for this module */

/* Why do I need to  provide callback if i dont use it :-S */
void dummy ( void ) { return;  }


extern ClCmContextT gClCmChassisMgrContext;
extern ClCmContextT gClHpiChassisMgrContext;

static void clCmEventDeliverCallback(SaEvtSubscriptionIdT subscriptionId, SaEvtEventHandleT eventHandle, SaSizeT eventDataSize)
{
    SaAisErrorT     rc = SA_AIS_OK;
    //static ClPtrT   resTest = 0;
    ClEventT * resTest;
    SaEvtEventPatternArrayT patternArray= {0};
    SaEvtEventPriorityT priority=0;
    SaTimeT retentionTime=0;
    SaNameT publisherName={0};
    SaTimeT publishTime=0;
    SaEvtEventIdT eventId=0;
    SaSizeT     res_size = 0;

    clLog(CL_LOG_SEV_DEBUG, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Received an Event");
    clLog(CL_LOG_SEV_DEBUG, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Subscription ID  :ox%X",subscriptionId);
    clLog(CL_LOG_SEV_DEBUG, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "eventDataSize :[%llu]",eventDataSize);

    /* A high performance implementation would keep the buffer
       if it was big enough for the next event, OR even faster
       preallocate buffer(s) of the the maximum event size which
       can be known by controlling what processes publish to
       /   a particular event channel.

       This tutorial will simply free an realloc the event buffer.
     */
    resTest =(ClEventT *)clHeapAllocate(sizeof(ClEventT)+sizeof(ClCmSensorEventInfoT));
    if (0 == resTest) {
        clLog(CL_LOG_SEV_ERROR, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Failed to allocate space for event");
        return;
    }

    rc=saEvtEventAttributesGet(eventHandle,&patternArray,&priority,&retentionTime,&publisherName,&publishTime,&eventId);
    /* This API can be used outside of the callback routine, which is why
       you need to pass the size of the buffer you've allocated. */

    if (rc!= SA_AIS_OK)
    {
        clLog(CL_LOG_SEV_DEBUG, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Failed to get event data attributes [0x%x]",rc);
        return;
    }

    else
    {
        clLog(CL_LOG_SEV_DEBUG, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "EVENT ID        : 0x%llu",eventId);
        clLog(CL_LOG_SEV_DEBUG, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "EVENT priority  : 0x%X",priority);    
        clLog(CL_LOG_SEV_DEBUG, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Retention time  : 0x%llu",retentionTime);

    }
    res_size=sizeof(ClEventT)+sizeof(ClCmSensorEventInfoT);
    rc = saEvtEventDataGet(eventHandle, resTest, &res_size);
    if (rc!= SA_AIS_OK)
    {
        clLog(CL_LOG_SEV_ERROR, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Failed to get event data  [0x%x]",rc);
        return;
    }


    // clprintf(CL_LOG_SEV_INFO,"Ani Data Size: %lu\n",eventDataSize);



    /* This tutorial just prints the event.  Rather then rely on the correct
       null termination at the publisher side, I'm going to append a 0 at
       the end before printing it. */
    *(((char *)resTest) + eventDataSize) = 0;
    //    clprintf(CL_LOG_SEV_INFO,"=============received event data start============\n");

    clLog(CL_LOG_SEV_DEBUG, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Event TYPE :[%x]",resTest->PayloadType);       
    clLog(CL_LOG_SEV_DEBUG, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "=====received event data end=======\n");

}

/* Called by Chassis manager to Initialize the Event Interface */
ClRcT clCmEventInterfaceInitialize(void )
{

    ClRcT rc = CL_OK;
    ClVersionT eventLibVersion = CL_EVENT_VERSION;
    ClEventCallbacksT callbacks = {
        (ClEventChannelOpenCallbackT) dummy, 
        (ClEventDeliverCallbackT)     clCmEventDeliverCallback 
    };
    ClNameT cmEvtChannelName = {  
        /*strlen doen not includes '\0'*/
        (ClUint16T)strlen(CL_CM_EVENT_CHANNEL),
        CL_CM_EVENT_CHANNEL 
    };
    ClNameT HpiEvtChannelName = {
        (ClUint16T)strlen(HPI_EVENT_CHANNEL),
        HPI_EVENT_CHANNEL
    };

    memset((void*)&gClCmChassisMgrContext.eventInfo.EventLibHandle,0, sizeof(ClEventHandleT ));
    memset((void*)&gClCmChassisMgrContext.eventInfo.EventChannelHandle, 0, sizeof(ClEventChannelHandleT));
    memset((void*)&gClHpiChassisMgrContext.eventInfo.EventLibHandle, 0, sizeof(ClEventHandleT ));
    memset((void*)&gClHpiChassisMgrContext.eventInfo.EventChannelHandle, 0, sizeof(ClEventChannelHandleT));

    rc = clEventInitialize( &gClCmChassisMgrContext.eventInfo.EventLibHandle, &callbacks, &eventLibVersion);
    if ( rc != CL_EVENT_ERR_ALREADY_INITIALIZED && rc != CL_OK )
    {
        clLog(CL_LOG_SEV_ERROR, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Event client library initialization failed for CM channel, error [0x%x]", rc);
        return rc;
    }

    gClHpiChassisMgrContext.eventInfo.EventLibHandle = gClCmChassisMgrContext.eventInfo.EventLibHandle;

    rc = clEventChannelOpen( gClCmChassisMgrContext.eventInfo.EventLibHandle, &cmEvtChannelName, 
            (ClEventChannelOpenFlagsT) (CL_EVENT_CHANNEL_CREATE | CL_EVENT_CHANNEL_PUBLISHER | CL_EVENT_GLOBAL_CHANNEL),
            -1, &gClCmChassisMgrContext.eventInfo.EventChannelHandle );

    if ( rc != CL_EVENT_ERR_EXIST && rc != CL_OK )
    {
        clLog(CL_LOG_SEV_ERROR, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Could not open cm event channel for CM, error [0x%x]", rc);
        clEventFinalize(gClCmChassisMgrContext.eventInfo.EventLibHandle);
        return rc;
    }

    rc = clEventChannelOpen( gClHpiChassisMgrContext.eventInfo.EventLibHandle, &HpiEvtChannelName,
            (ClEventChannelOpenFlagsT) (CL_EVENT_CHANNEL_CREATE | CL_EVENT_CHANNEL_PUBLISHER | CL_EVENT_GLOBAL_CHANNEL),
            -1, &gClHpiChassisMgrContext.eventInfo.EventChannelHandle );

    if ( rc != CL_EVENT_ERR_EXIST && rc != CL_OK )
    {
        clLog(CL_LOG_SEV_ERROR, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Could not open hpi event channel for CM, error [0x%x]", rc);
        clEventFinalize(gClHpiChassisMgrContext.eventInfo.EventLibHandle);
        return rc;
    }



    /* Allocate the event */
    memset((void*)&gClCmChassisMgrContext.eventInfo.EvtHandle, 0, sizeof(ClEventHandleT));
    memset((void*)&gClHpiChassisMgrContext.eventInfo.EvtHandle, 0, sizeof(ClEventHandleT));

    if ((rc=clEventAllocate(gClCmChassisMgrContext.eventInfo.EventChannelHandle, &gClCmChassisMgrContext.eventInfo.EvtHandle ))!= CL_OK)
    {
        clLog(CL_LOG_SEV_ERROR, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Event allocation failed, error [0x%x]", rc);
        clEventChannelClose(gClCmChassisMgrContext.eventInfo.EventChannelHandle);
        clEventFinalize(gClCmChassisMgrContext.eventInfo.EventLibHandle);
        return rc;
    }

    if ((rc=clEventAllocate(gClHpiChassisMgrContext.eventInfo.EventChannelHandle, &gClHpiChassisMgrContext.eventInfo.EvtHandle ))!= CL_OK)
    {
        clLog(CL_LOG_SEV_ERROR, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Event allocation failed, error [0x%x]", rc);
        clEventChannelClose(gClHpiChassisMgrContext.eventInfo.EventChannelHandle);
        clEventFinalize(gClHpiChassisMgrContext.eventInfo.EventLibHandle);
        return rc;
    }
    gClCmChassisMgrContext.eventInfo.eventIfaceInitDone = CL_TRUE;
    gClHpiChassisMgrContext.eventInfo.eventIfaceInitDone = CL_TRUE;
    clLog(CL_LOG_SEV_DEBUG, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "CM Event Interface is Successfully Initialized");

    return rc;
}


/* Called by Chassis manager to Finalize the Event Interface */
ClRcT clCmEventInterfaceFinalize( void )
{

    ClRcT rc = CL_OK;

    if( gClCmChassisMgrContext.eventInfo.eventIfaceInitDone )
    {
        clEventFree( gClCmChassisMgrContext.eventInfo.EvtHandle );
        clEventChannelClose( gClCmChassisMgrContext.eventInfo.EventChannelHandle );
        clEventFinalize( gClCmChassisMgrContext.eventInfo.EventLibHandle );
    }

    if( gClHpiChassisMgrContext.eventInfo.eventIfaceInitDone )
    {
        clEventFree( gClHpiChassisMgrContext.eventInfo.EvtHandle );
        clEventChannelClose( gClHpiChassisMgrContext.eventInfo.EventChannelHandle );
        clEventFinalize( gClHpiChassisMgrContext.eventInfo.EventLibHandle );
    }


    gClCmChassisMgrContext.eventInfo.eventIfaceInitDone = CL_FALSE;
    gClHpiChassisMgrContext.eventInfo.eventIfaceInitDone = CL_FALSE;
    clLog(CL_LOG_SEV_DEBUG, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "CM Event channel closed");
    clLog(CL_LOG_SEV_DEBUG, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "HPI Event channel closed");

    return rc;
}

ClRcT clCmHpiPublishEvent( CmEventIdT cmEventId, ClUint32T payLoadLen, void *payload )
{
    ClRcT rc = CL_OK;
    ClEventIdT eventId; 
    ClNameT chassisManagerEvent ;
    ClEventPatternT evtPattern ;
    ClEventPatternArrayT evtPatternArray;

    memset(&chassisManagerEvent, 0, sizeof(ClNameT));

    if (CM_FRU_STATE_TRANSITION_EVENT == cmEventId)
    {   
        clNameSet(&chassisManagerEvent,(const char *)CL_CM_FRU_STATE_TRANSITION_EVENT_STR);
    }
    else if (CM_ALARM_EVENT == cmEventId) {
        clNameSet(&chassisManagerEvent,(const char *)CL_CM_ALARM_EVENT_STR);
    }
    else
    {
        clLog(CL_LOG_SEV_ERROR, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Invalid event type [%u]", cmEventId); return CL_CM_ERROR(CL_ERR_INVALID_PARAMETER);
    }


    evtPattern.allocatedSize = 0;
    evtPattern.patternSize = (ClSizeT)chassisManagerEvent.length + 1;
    evtPattern.pPattern = (ClUint8T *)chassisManagerEvent.value;

    evtPatternArray.allocatedNumber = 0;
    evtPatternArray.patternsNumber = 1;
    evtPatternArray.pPatterns = &evtPattern;


    if((rc = clEventAttributesSet( gClHpiChassisMgrContext.eventInfo.EvtHandle, &evtPatternArray, 0x01, 0, &chassisManagerEvent))!= CL_OK )
    {
        clLog(CL_LOG_SEV_ERROR, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Failed to set event attributes, error [0x%x]", rc);
        return rc;
    }

    if((rc = clEventPublish( gClHpiChassisMgrContext.eventInfo.EvtHandle, (void*)payload, payLoadLen, &eventId))!= CL_OK) 
    {
        clLog(CL_LOG_SEV_ERROR, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Could not publish event, error [0x%x]", rc);
        return rc;
    }
    clLog(CL_LOG_SEV_DEBUG, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Event published");
    return rc;
}

ClRcT clCmPublishEvent( CmEventIdT cmEventId, ClUint32T payLoadLen, void *payload )
{
    ClRcT rc = CL_OK;
    ClEventIdT eventId; 
    ClNameT chassisManagerEvent ;
    ClEventPatternT evtPattern ;
    ClEventPatternArrayT evtPatternArray;

    if(!gClCmChassisMgrContext.eventInfo.eventIfaceInitDone)
    {
        clLog(CL_LOG_SEV_NOTICE, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED,
              "Event channel already finalized as CM appears to be shutting down. "
              "Skipping event type [%u]", cmEventId);
        return CL_OK;
    }

    memset(&chassisManagerEvent, 0, sizeof(ClNameT));

    if (CM_FRU_STATE_TRANSITION_EVENT == cmEventId)
    {
        clNameSet(&chassisManagerEvent,(const char *)CL_CM_FRU_STATE_TRANSITION_EVENT_STR);
    }
    else if (CM_ALARM_EVENT == cmEventId) {
        clNameSet(&chassisManagerEvent,(const char *)CL_CM_ALARM_EVENT_STR);
    }
    else{
        clLog(CL_LOG_SEV_ERROR, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Invalid event type [%u]", cmEventId);
        return CL_CM_ERROR(CL_ERR_INVALID_PARAMETER);
    }


    evtPattern.allocatedSize = 0;
    evtPattern.patternSize = (ClSizeT)chassisManagerEvent.length + 1;
    evtPattern.pPattern = (ClUint8T *)chassisManagerEvent.value;

    evtPatternArray.allocatedNumber = 0;
    evtPatternArray.patternsNumber = 1;
    evtPatternArray.pPatterns = &evtPattern;


    if((rc = clEventAttributesSet( gClCmChassisMgrContext.eventInfo.EvtHandle, &evtPatternArray, 0x01, 0, &chassisManagerEvent))!= CL_OK )
    {
        clLog(CL_LOG_SEV_ERROR, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Failed to set event attributes, error [0x%x]", rc);
        return rc;
    }

    if((rc = clEventPublish( gClCmChassisMgrContext.eventInfo.EvtHandle, (void*)payload, payLoadLen, &eventId))!= CL_OK)
    {
        clLog(CL_LOG_SEV_ERROR, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Could not publish event, error [0x%x]", rc);
        return rc;
    }
    clLog(CL_LOG_SEV_DEBUG, AREA_EVT, CL_LOG_CONTEXT_UNSPECIFIED, "Event published");
    return rc;
}


